package djmil.cordacheckers.checkers;

import static java.lang.Math.abs;

import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

public class Move {
    final public Integer from;
    final public Integer to;
    final public Integer jumpOver;

    public Move(int from, int to) {
        this.from = from;
        this.to = to;

        Set<Integer> fromSteps = getSteps(from);
        if (fromSteps.contains(to))
            this.jumpOver = null;
        else
            this.jumpOver = intersection(fromSteps, getSteps(to));
    }

    public static Stone.Color apply(List<Integer> moves, Map<Integer, Stone> board, Stone.Color moveColor) {
        Move move = new Move(moves.get(0), moves.get(1));
        return move.apply(board, moveColor);
    }

    private Stone.Color apply(Map<Integer, Stone> board, Stone.Color moveColor) {
        final Set<Move> mandatoryJumps = getPossibleJumps(board, moveColor);
        if (mandatoryJumps.size() != 0 && !mandatoryJumps.contains(this))
            throw new Exception("A mandatory opponent's stone capture must be performed");

        Stone movingStone = board.remove(from);
        if (movingStone == null)
            throw new Exception("An empty starting tile");

        if (movingStone.getColor() != moveColor)
            throw new Exception("Only " +moveColor.name() +" color is expected to move");

        final Set<Move> allMoves = movingStone.getMoves(from);
        if (!allMoves.contains(this))
            throw new Exception("Prohibited move");

        movingStone = movingStone.promoteIfPossible(to);
        if (board.put(to, movingStone) != null)
            throw new Exception("An occupied finishing tile");

        if (isJump()) {
            final Stone jumpOver = board.remove(this.jumpOver);
            if (jumpOver == null || jumpOver.getColor() != moveColor.opposite())
                throw new Exception("Must jump over an opponent's stone");
            
            final Set<Move> chainJumps = movingStone.getJumps(this.to);
            chainJumps.removeIf(move -> !move.canJump(board, moveColor));
            if (chainJumps.size() != 0)
                return moveColor;       // player must continue his turn
        }

        return moveColor.opposite();    // player has finished his turn
    }

    boolean isJump() {
        return this.jumpOver != null;
    }

    boolean canJump(Map<Integer, Stone> board, Stone.Color color) {
        if (!isJump())
            return false;

        final Stone jumpStone = board.get(this.from);
        if (jumpStone == null || jumpStone.getColor() != color) 
            return false;
        
        final Stone jumpOverStone = board.get(this.jumpOver);
        if (jumpOverStone == null || jumpOverStone.getColor() != color.opposite())
            return false;
        
        if (board.get(this.to) != null)
            return false;
        
        return true;
    }

    boolean canMove(Map<Integer, Stone> board, Stone.Color color) {
        if (isJump())
            return canJump(board, color);

        final Stone stepStone = board.get(this.from);
        if (stepStone == null || stepStone.getColor() != color) 
            return false;
        
        if (board.get(this.to) != null)
            return false;
        
        return true;
    }

    static Integer intersection(Set<Integer> a, Set<Integer> b) {
        Set<Integer> intersection = new HashSet<Integer>(a);

        intersection.retainAll(b);
        if (intersection.size() != 1)
            // A legit move is characterized by a single intersection point
            throw new Exception("Prohibited move");
        
        return intersection.iterator().next();
    }

    static Set<Integer> getSteps(Integer idx) {
        return adjacentCell.get(idx).stream()
            .filter(cur -> abs(idx - cur) <= 5)
            .collect(Collectors.toSet());
    }

    static Set<Integer> getJumps(Integer idx) {
        return adjacentCell.get(idx).stream()
            .filter(cur -> abs(idx - cur) > 5)
            .collect(Collectors.toSet());
    }

    static Set<Integer> getBlackSteps(Integer idx) {
        return adjacentCell.get(idx).stream()
            .filter(cur -> cur > idx && cur - idx <= 5)
            .collect(Collectors.toSet());
    }

    static Set<Integer> getWhiteSteps(Integer idx) {
        return adjacentCell.get(idx).stream()
            .filter(cur -> cur < idx && idx - cur <= 5)
            .collect(Collectors.toSet());
    }

    static Set<Integer> getBlackJumps(Integer idx) {
        return adjacentCell.get(idx).stream()
            .filter(cur -> cur > idx && cur - idx > 5)
            .collect(Collectors.toSet());
    }

    static Set<Integer> getWhiteJumps(Integer idx) {
        return adjacentCell.get(idx).stream()
            .filter(cur -> cur < idx && idx - cur > 5)
            .collect(Collectors.toSet());
    }

    final static Map<Integer, List<Integer>> adjacentCell = Map.ofEntries(
        Map.entry(1, Arrays.asList(5, 6, 10)),
        Map.entry(2, Arrays.asList(6, 7, 9, 11)),
        Map.entry(3, Arrays.asList(7, 8, 10, 12)),
        Map.entry(4, Arrays.asList(8, 11)),
        Map.entry(5, Arrays.asList(1, 9, 14)),
        Map.entry(6, Arrays.asList(1, 2, 9, 10, 13, 15)),
        Map.entry(7, Arrays.asList(2, 3, 10, 11, 14, 16)),
        Map.entry(8, Arrays.asList(3, 4, 11, 12, 15)),
        Map.entry(9, Arrays.asList(2, 5, 6, 13, 14, 18)),
        Map.entry(10, Arrays.asList(1, 3, 6, 7, 14, 15, 17, 19)),
        Map.entry(11, Arrays.asList(2, 4, 7, 8, 15, 16, 18, 20)),
        Map.entry(12, Arrays.asList(3, 8, 16, 19)),
        Map.entry(13, Arrays.asList(6, 9, 17, 22)),
        Map.entry(14, Arrays.asList(5, 7, 9, 10, 17, 18, 21, 23)),
        Map.entry(15, Arrays.asList(6, 8, 10, 11, 18, 19, 22, 24)),
        Map.entry(16, Arrays.asList(7, 11, 12, 19, 20, 23)),
        Map.entry(17, Arrays.asList(10, 13, 14, 21, 22, 26)),
        Map.entry(18, Arrays.asList(9, 11, 14, 15, 22, 23, 25, 27)),
        Map.entry(19, Arrays.asList(10, 12, 15, 16, 23, 24)),
        Map.entry(20, Arrays.asList(11, 16, 24, 27)),
        Map.entry(21, Arrays.asList(14, 17, 25, 30)),
        Map.entry(22, Arrays.asList(13, 15, 17, 18, 25, 26, 29, 31)),
        Map.entry(23, Arrays.asList(14, 16, 18, 19, 26, 27, 30, 32)),
        Map.entry(24, Arrays.asList(15, 19, 20, 27, 28, 31)),
        Map.entry(25, Arrays.asList(18, 21, 22, 29, 30)),
        Map.entry(26, Arrays.asList(17, 19, 22, 23, 30, 31)),
        Map.entry(27, Arrays.asList(18, 20, 23, 24, 31, 32)),
        Map.entry(28, Arrays.asList(19, 24, 32)),
        Map.entry(29, Arrays.asList(22, 25)),
        Map.entry(30, Arrays.asList(21, 23, 25, 26)),
        Map.entry(31, Arrays.asList(22, 24, 26, 27)),
        Map.entry(32, Arrays.asList(23, 27, 28))
    );

    public static class Exception extends RuntimeException {
        public Exception(String message) {
            super(message);
        }
    }

    static Set<Move> getPossibleJumps(Map<Integer, Stone> board, Stone.Color color) {
        Set<Move> res = new HashSet<Move>();

        for (Map.Entry<Integer, Stone> entry : board.entrySet()) {
            final Integer from = entry.getKey();
            final Stone  stone = entry.getValue();

            if (stone.getColor() != color)
                continue;

            final Set<Move> jumps = stone.getJumps(from);
            jumps.removeIf( move -> !move.canJump(board, color));
            res.addAll(jumps);
        }

        return res;
    }

    public static Set<Move> getPossibleMoves(Map<Integer, Stone> board, Stone.Color color) {
        Set<Move> res = new HashSet<Move>();

        for (Map.Entry<Integer, Stone> entry : board.entrySet()) {
            final Integer from = entry.getKey();
            final Stone  stone = entry.getValue();

            if (stone.getColor() != color)
                continue;

            final Set<Move> moves = stone.getMoves(from);
            moves.removeIf( move -> !move.canMove(board, color));
            res.addAll(moves);
        }

        return res;
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((from == null) ? 0 : from.hashCode());
        result = prime * result + ((to == null) ? 0 : to.hashCode());
        result = prime * result + ((jumpOver == null) ? 0 : jumpOver.hashCode());
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        Move other = (Move) obj;
        if (from == null) {
            if (other.from != null)
                return false;
        } else if (!from.equals(other.from))
            return false;
        if (to == null) {
            if (other.to != null)
                return false;
        } else if (!to.equals(other.to))
            return false;
        if (jumpOver == null) {
            if (other.jumpOver != null)
                return false;
        } else if (!jumpOver.equals(other.jumpOver))
            return false;
        return true;
    }

    @Override
    public String toString() {
        return "[from=" + from + ", to=" + to + ", jumpOver=" + jumpOver + "]";
    }

}
